package ch.unige.solidify.controller;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.linkTo;

import java.net.URI;
import java.util.List;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.FileSystemResource;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.hateoas.RepresentationModel;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.ResponseBody;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.Collection;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.rest.SubResource;
import ch.unige.solidify.rest.SubResourceContainer;
import ch.unige.solidify.rest.Tool;
import ch.unige.solidify.security.NoPermissions;
import ch.unige.solidify.service.CompositeResourceService;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.SearchCriteria;

/**
 * A controller for subresources
 *
 * @param <T>
 *          the type of the main resource
 * @param <V>
 *          the type of the subresource
 */
public abstract class SubResourceController<T extends Resource<T> & SubResourceContainer, V extends Resource<V>>
        extends SolidifyController {

  @Autowired
  protected CompositeResourceService<T> itemService;

  @Autowired
  private CompositeResourceService<V> subResourceService;

  @NoPermissions
  @GetMapping(value = SolidifyConstants.URL_ID_PLUS_SEP + ActionName.DL)
  @ResponseBody
  public HttpEntity<FileSystemResource> download(@PathVariable String parentid,
          @PathVariable String id) {
    this.checkIfResourceExists(parentid);
    final V item = this.itemService
            .findByParentIdAndChildId(parentid, this.getMainResourceClass(), this.getSubResourceClass(), id);
    if (item == null) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
    final URI fileUri = item.getDownloadUri();
    return this.buildDownloadResponseEntity(fileUri.getPath(), fileUri.getPath().substring(fileUri.getPath().lastIndexOf('/') + 1),
            FileTool.getContentType(fileUri).toString(), FileTool.getSize(fileUri));
  }

  @GetMapping(value = SolidifyConstants.URL_ID)
  public HttpEntity<V> get(@PathVariable String parentid, @PathVariable String id) {
    this.checkIfResourceExists(parentid);
    final V item = this.itemService
            .findByParentIdAndChildId(parentid, this.getMainResourceClass(), this.getSubResourceClass(), id);
    if (item != null) {
      this.addLinks(parentid, item);
      return new ResponseEntity<>(item, HttpStatus.OK);
    }
    return new ResponseEntity<>(HttpStatus.NOT_FOUND);
  }

  @GetMapping
  public HttpEntity<Collection<V>> list(@PathVariable String parentid, @ModelAttribute V filterItem,
          Pageable pageable) {

    this.checkIfResourceExists(parentid);

    Page<V> list;
    if (filterItem instanceof SubResource) {
      ((SubResource) filterItem).setParentResourceId(parentid);
      list = this.subResourceService.findAll(filterItem.getSpecification(), pageable);
    } else {
      if (pageable.getSort().isSorted()) {
        throw new UnsupportedOperationException(this.messageService.get("error.sort.unsupported",
                new Object[] { this.getMainResourceClass().toString(),
                        this.getSubResourceClass().toString() }));
      }
      list = this.itemService
              .findByParentId(parentid, this.getMainResourceClass(), this.getSubResourceClass(), pageable);
    }

    for (final V v : list) {
      this.addLinks(parentid, v);
    }
    final Collection<V> collection = this.processPageAndLinks(list, parentid, pageable);
    return new ResponseEntity<>(collection, HttpStatus.OK);
  }

  public HttpEntity<Collection<V>> listBySubResourceSearchCriteria(SearchCriteria parent, V filterItem, String orOpredicateFlag,
          List<SearchCriteria> criterias, Pageable pageable) {
    final String parentid = parent.getValue().toString();
    this.checkIfResourceExists(parentid);
    final Page<V> list = this.subResourceService.findBySubResourceSearchCriteria(filterItem, orOpredicateFlag, criterias, parent, pageable);
    for (final V v : list) {
      this.addLinks(parentid, v);
    }
    final Collection<V> collection = this.processPageAndLinks(list, parentid, pageable);
    return new ResponseEntity<>(collection, HttpStatus.OK);
  }

  protected abstract void addLinks(String parentid, V v);

  protected <W extends RepresentationModel<W>> void addOthersLinks(String parentid, W w) {

  }

  protected void checkIfResourceExists(String id) {
    // if resource doen't exist throw a NoSuchElementElement catched by SolidifyControllerAdvice
    this.itemService.findOne(id);
  }

  protected abstract Class<T> getMainResourceClass();

  protected abstract Class<V> getSubResourceClass();

  protected Collection<V> processPageAndLinks(Page<V> list, String parentid, Pageable pageable) {
    final Collection<V> collection = new Collection<>(list, pageable);
    collection.add(linkTo(this.getClass(), parentid).withSelfRel());
    collection.add((Tool.parentLink((linkTo(this.getClass(), parentid)).toUriComponentsBuilder()))
            .withRel(ActionName.PARENT));
    this.addSortLinks(linkTo(this.getClass(), parentid, pageable), collection);
    this.addPageLinks(linkTo(this.getClass(), parentid), collection, pageable);
    this.addOthersLinks(parentid, collection);
    return collection;
  }
}
