package ch.unige.solidify.service;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.data.domain.Sort.Order;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpMethod;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.RestTemplate;
import org.springframework.web.util.UriComponentsBuilder;

import ch.unige.solidify.exception.RestResponseErrorHandler;
import ch.unige.solidify.exception.SolidifyResourceNotFoundException;
import ch.unige.solidify.exception.SolidifyRestException;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.Collection;
import ch.unige.solidify.rest.CollectionPage;
import ch.unige.solidify.rest.Result;
import ch.unige.solidify.util.CollectionTool;
import ch.unige.solidify.util.RestClientTool;

public abstract class SolidifyRestClientService {
  private static final Logger log = LoggerFactory.getLogger(SolidifyRestClientService.class);

  protected MessageService messageService;
  private final AuthenticationStorageProxy authenticationStorageProxy;
  private final RestClientTool restClientTool;

  public SolidifyRestClientService(RestClientTool restClientTool, MessageService messageService,
          AuthenticationStorageProxy authenticationStorageProxy) {
    this.restClientTool = restClientTool;
    this.messageService = messageService;
    this.authenticationStorageProxy = authenticationStorageProxy;
  }

  /**
   * @param <T>
   * @param url
   * @param clazz
   * @param pageable
   * @return
   */
  public <T> Collection<T> getResourceList(String url, Class<T> clazz, Pageable pageable) {
    final String jsonResult = this.getResources(url, pageable);
    final List<T> list = CollectionTool.getList(jsonResult, clazz);
    final CollectionPage page = CollectionTool.getPage(jsonResult);
    return new Collection<>(list, page);
  }

  /**
   * Complete an URL with pagination query parameters
   * 
   * @param url
   *          url to complete
   * @param pageable
   *          pageable object used to complete the URL
   * @return completed URL
   */
  protected String completeUrlWithPagination(String url, Pageable pageable) {
    final UriComponentsBuilder uriBuilder = UriComponentsBuilder.fromUriString(url);

    if (pageable != null) {
      final MultiValueMap<String, String> existingQueryParams = new LinkedMultiValueMap(uriBuilder.build().getQueryParams());
      existingQueryParams.remove("page");
      existingQueryParams.remove("size");
      existingQueryParams.remove("sort");

      final int page = pageable.getPageNumber();
      final int size = pageable.getPageSize();
      final Sort sort = pageable.getSort();

      final MultiValueMap<String, String> pageParams = new LinkedMultiValueMap<>();
      pageParams.set("page", Integer.toString(page));
      pageParams.set("size", Integer.toString(size));

      // TODO: support sorting on many fields ?
      for (final Order order : sort) {
        final String sortProperty = order.getProperty();
        final String sortDirection = order.getDirection().toString();
        pageParams.set("sort", sortProperty + "," + sortDirection);
      }
      uriBuilder.replaceQueryParams(existingQueryParams).queryParams(pageParams);
    }
    return uriBuilder.build().toString();
  }

  protected void deleteResource(String url) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.DELETE);
    final HttpEntity<?> httpEntity = new HttpEntity<>(this.getAuthorizationHeaders());
    try {
      restClt.exchange(url, HttpMethod.DELETE, httpEntity, Void.class);
    } catch (final HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      }
      final String message = this.getResourceErrorMessage(url);
      throw new SolidifyRestException(message, e);
    }
  }

  protected MultiValueMap<String, String> getAuthorizationHeaders() {
    if (this.authenticationStorageProxy.getAuthenticationStorage() != null) {
      return this.authenticationStorageProxy.getAuthenticationStorage()
              .getHeadersWithAuthorization();
    }
    return new LinkedMultiValueMap<>();
  }

  protected String getResource(String url) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.GET);
    final HttpEntity<?> httpEntity = new HttpEntity<>(this.getAuthorizationHeaders());
    try {
      final ResponseEntity<String> responseEntity = restClt.exchange(url, HttpMethod.GET, httpEntity, String.class);
      return responseEntity.getBody();
    } catch (final HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      }
      final String message = this.getResourceErrorMessage(url);
      throw new SolidifyRestException(message, e);
    }
  }

  protected <T> T getResource(String url, Class<T> clazz) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.GET);
    final HttpEntity<?> httpEntity = new HttpEntity<>(this.getAuthorizationHeaders());
    try {
      final ResponseEntity<T> responseEntity = restClt.exchange(url, HttpMethod.GET, httpEntity, clazz);
      return responseEntity.getBody();
    } catch (final HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      }
      final String message = this.getResourceErrorMessage(url);
      throw new SolidifyRestException(message, e);
    }
  }

  protected String getResource(String url, Pageable pageable) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.GET);
    final HttpEntity<?> httpEntity = new HttpEntity<>(this.getAuthorizationHeaders());
    final StringBuilder urlBuilder = new StringBuilder(url);
    // Add Pagination info
    if (pageable != null) {
      if (url.contains("?")) {
        urlBuilder.append("&");
      } else {
        urlBuilder.append("?");
      }
      urlBuilder.append(ActionName.PAGE + "=" + pageable.getPageNumber() + "&" + ActionName.SIZE + "=" + pageable.getPageSize());
      // Add Sort info
      pageable.getSort()
              .stream()
              .forEach(s -> urlBuilder.append("&" + ActionName.SORT + "=" + s.getProperty() + "," + s.getDirection().toString().toLowerCase()));
    }
    url = urlBuilder.toString();
    try {
      this.logModuleCall(url);
      final ResponseEntity<String> responseEntity = restClt.exchange(url, HttpMethod.GET, httpEntity, String.class);
      return responseEntity.getBody();
    } catch (final HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      }
      final String message = this.getResourceErrorMessage(url);
      throw new SolidifyRestException(message, e);
    }
  }

  // ********************
  // ** HEAD Functions **
  // ********************

  protected String getResources(String url, Pageable pageable) {
    url = this.completeUrlWithPagination(url, pageable);

    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.GET);
    final HttpEntity<?> httpEntity = new HttpEntity<>(this.getAuthorizationHeaders());
    try {

      this.logModuleCall(url);

      return restClt.exchange(url, HttpMethod.GET, httpEntity, String.class).getBody();
    } catch (final HttpClientErrorException e) {

      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      } else {
        final String message = this.getResourceErrorMessage(url);
        throw new SolidifyRestException(message, e);
      }
    }
  }

  // ********************
  // ** POST Functions **
  // ********************

  // *******************
  // ** GET Functions **
  // *******************
  protected RestTemplate getRestTemplate(String url, HttpMethod httpMethod) {
    final RestTemplate restClt = this.restClientTool.getClient(url);
    switch (httpMethod) {
      case GET:
      case HEAD:
        break;
      default:
        restClt.setErrorHandler(new RestResponseErrorHandler());
    }
    return restClt;
  }

  protected HttpStatus headResource(String url) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.HEAD);
    final HttpEntity<?> httpEntity = new HttpEntity<>(this.getAuthorizationHeaders());
    try {
      restClt.exchange(url, HttpMethod.HEAD, httpEntity, Void.class);
    } catch (final HttpClientErrorException e) {
      return e.getStatusCode();
    }
    return HttpStatus.OK;
  }

  protected <T> T patchResource(String url, T updatedResource, Class<T> clazz) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.PATCH);
    restClt.setErrorHandler(new RestResponseErrorHandler());
    final HttpEntity<?> httpEntity = new HttpEntity<>(updatedResource, this.getAuthorizationHeaders());
    try {
      return restClt.exchange(url, HttpMethod.PATCH, httpEntity, clazz).getBody();
    } catch (final HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      }
      final String message = this.getResourceErrorMessage(url);
      throw new SolidifyRestException(message, e);
    }
  }

  protected <T> T postResource(String url, Class<T> clazz) {
    final RestTemplate restClt = this.restClientTool.getClient(url);
    final HttpEntity<?> httpEntity = new HttpEntity<>(this.getAuthorizationHeaders());
    try {

      final ResponseEntity<T> responseEntity = restClt.exchange(url, HttpMethod.POST, httpEntity, clazz);
      return responseEntity.getBody();

    } catch (final HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      }
      final String message = this.getResourceErrorMessage(url);
      throw new SolidifyRestException(message, e);
    }
  }

  // **********************
  // ** DELETE Functions **
  // **********************

  protected <T> T postResource(String url, T newResource, Class<T> clazz) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.POST);
    final HttpEntity<?> httpEntity = new HttpEntity<>(newResource, this.getAuthorizationHeaders());
    try {
      final ResponseEntity<T> responseEntity = restClt.exchange(url, HttpMethod.POST, httpEntity, clazz);
      return responseEntity.getBody();
    } catch (final HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      }
      final String message = this.getResourceErrorMessage(url);
      throw new SolidifyRestException(message, e);
    }
  }

  // ********************
  // ** PATCH Function **
  // ********************

  protected Result postResourceAction(String url) {
    return this.postResource(url, Result.class);
  }

  // *********************
  // ** Misc. Functions **
  // *********************

  protected <T, V> T postResourceOther(String url, V otherResource, Class<T> clazz) {
    final RestTemplate restClt = this.getRestTemplate(url, HttpMethod.POST);
    final HttpEntity<?> httpEntity = new HttpEntity<>(otherResource, this.getAuthorizationHeaders());
    try {
      final ResponseEntity<T> responseEntity = restClt.exchange(url, HttpMethod.POST, httpEntity, clazz);
      return responseEntity.getBody();
    } catch (final HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      }
      final String message = this.getResourceErrorMessage(url);
      throw new SolidifyRestException(message, e);
    }
  }

  private String getResourceErrorMessage(String url) {
    return "Error in getting resource (" + url + ")";
  }

  private void logModuleCall(String moduleUrl) {
    try {
      final URL url = new URL(moduleUrl);
      final String userInfo = url.getUserInfo();
      if (userInfo != null) {
        final String userInfoWithoutPassword = userInfo.substring(0, userInfo.indexOf(':')) + ":**********";
        final String urlWithoutPassword = url.getProtocol() + "://" + userInfoWithoutPassword + "@" + url.getHost() + url
                .getPort() + url.getPath() + "?" + url.getQuery();
        log.info("Module call GET: {}", urlWithoutPassword);
      } else {
        log.info("Module call GET: {}", moduleUrl);
      }
    } catch (final MalformedURLException e) {
      log.error("Malformed url : {}", moduleUrl, e);
    }
  }

}
